#      A python library for getting Load Shedding schedules from Eskom.
#      Copyright (C) 2021  Werner Pieterson
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import unittest
from unittest import TestCase

from .eskom import Eskom, AreaInfo, Province, Stage, Suburb, Municipality

skip_slow = os.environ.get('skip_slow', False)


@unittest.skipIf(skip_slow, "Slow (Network calls)")
class TestEskom(TestCase):

    def setUp(self) -> None:
        self.eskom = Eskom()

    def test_find_suburbs(self):
        suburbs = self.eskom.find_suburbs(search_text="Milnerton")
        self.assertEqual(list, type(suburbs))
        self.assertNotEqual(len(suburbs), 0)
        self.assertEqual(Suburb, type(suburbs[0]))
        self.assertEqual(1058852, suburbs[0].id)

    def test_get_municipalities(self):
        municipalities = self.eskom.get_municipalities(province=Province(id=Province.WESTERN_CAPE))
        self.assertEqual(list, type(municipalities))
        self.assertNotEqual(len(municipalities), 0)
        self.assertEqual(Municipality, type(municipalities[0]))
        self.assertEqual("Beaufort West", municipalities[0].name)

    def test_get_schedule(self):
        schedule = self.eskom.get_schedule(province=Province(id=Province.WESTERN_CAPE), suburb=Suburb(id=1058852),
                                           stage=Stage.STAGE_2)
        self.assertEqual(list, type(schedule))
        self.assertNotEqual(len(schedule), 0)
        self.assertNotEqual(len(schedule[0]), 0)

    def test_get_schedule_area_info(self):
        area_info = self.eskom.get_schedule_area_info(suburb=Suburb(id=1058852))
        self.assertEqual(AreaInfo, type(area_info))
        self.assertEqual(Province.WESTERN_CAPE, area_info.province.id)
        self.assertEqual("Milnerton", area_info.suburb.name)

    def test_get_stage(self):
        stage = self.eskom.get_stage()
        self.assertEqual(Stage, type(stage))
        self.assertEqual(Stage, type(stage))

    def test_search(self):
        suburbs = self.eskom.find_suburbs(search_text="Milnerton")
        schedule = self.eskom.get_schedule(province=suburbs[0].province, suburb=suburbs[0], stage=Stage.STAGE_2)
        self.assertEqual(list, type(schedule))
        self.assertNotEqual(len(schedule), 0)
        self.assertNotEqual(len(schedule[0]), 0)



#      A python library for getting Load Shedding schedules from Eskom.
#      Copyright (C) 2021  Werner Pieterson
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import json
import logging
from enum import Enum
from typing import List, Tuple, Any

import certifi
import urllib3
from bs4 import BeautifulSoup


class AreaInfo:
    def __init__(self, /, **kwargs):
        self.province = Province(**kwargs.get("province"))
        self.city = City(**kwargs.get("city"))
        self.suburb = Suburb(**kwargs.get("suburb"))
        self.period = kwargs.get("period")


class Suburb:
    def __init__(self, /, **kwargs):
        self.id = kwargs.get("Id", kwargs.get("id"))
        self.name = kwargs.get("Name", kwargs.get("name"))
        self.municipality = kwargs.get("MunicipalityName")
        self.province = Province(name=kwargs.get("ProvinceName"))
        self.total = kwargs.get("Total")

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<{name} ({id}), {municipality}, {province}>".format(
            id=self.id,
            name=self.name,
            municipality=self.municipality,
            province=self.province
        )


class Municipality:
    def __init__(self, /, **kwargs):
        self.id = kwargs.get("Value", kwargs.get("id"))
        self.name = kwargs.get("Text", kwargs.get("name"))
        self.selected = kwargs.get("Selected")
        self.group = kwargs.get("Group")
        self.disabled = kwargs.get("Disabled")


class City:
    def __init__(self, /, **kwargs):
        self.id = kwargs.get("Id", kwargs.get("id"))
        self.name = kwargs.get("Name", kwargs.get("name"))


class Province:
    EASTERN_CAPE = 1
    FREE_STATE = 2
    GAUTENG_ = 3
    KWAZULU_NATAL = 4
    LIMPOPO_ = 6
    MPUMALANGA = 5
    NORTH_WEST = 7
    NORTERN_CAPE = 8
    WESTERN_CAPE = 9

    def __init__(self, /, **kwargs):
        self.id = kwargs.get("Id", kwargs.get("id"))
        self.name = kwargs.get("Name", kwargs.get("name"))

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<{name} ({id})>".format(
            id=self.id,
            name=self.name,
        )


class Stage(Enum):
    UNKNOWN = -1
    NO_LOAD_SHEDDING = 0
    STAGE_1 = 1
    STAGE_2 = 2
    STAGE_3 = 3
    STAGE_4 = 4
    STAGE_6 = 6
    STAGE_5 = 5
    STAGE_7 = 7
    STAGE_8 = 8

    @staticmethod
    def from_status(status: str) -> Enum:
        status_stage = {
            "1": Stage.NO_LOAD_SHEDDING.value,
            "2": Stage.STAGE_1.value,
            "3": Stage.STAGE_2.value,
            "4": Stage.STAGE_3.value,
            "6": Stage.STAGE_4.value,
            "5": Stage.STAGE_6.value,
            "7": Stage.STAGE_5.value,
            "8": Stage.STAGE_7.value,
            "9": Stage.STAGE_8.value,
        }
        return Stage(status_stage.get(status, Stage.UNKNOWN))

    def __str__(self):
        stage_name = {
            self.NO_LOAD_SHEDDING.value: "No Load Shedding",
            self.STAGE_1.value: "Stage 1",
            self.STAGE_2.value: "Stage 2",
            self.STAGE_3.value: "Stage 3",
            self.STAGE_4.value: "Stage 4",
            self.STAGE_6.value: "Stage 6",
            self.STAGE_5.value: "Stage 5",
            self.STAGE_7.value: "Stage 7",
            self.STAGE_8.value: "Stage 8",
        }
        return stage_name.get(self.value, "Unknown")


class Eskom:
    def __init__(self):
        self.base_url = "https://loadshedding.eskom.co.za/LoadShedding"

    def find_suburbs(self, search_text: str, max_results: int = 300) -> List[Suburb]:
        url = "{base_url}/FindSuburbs?searchText={search_text}&maxResults={max_results}".format(
            base_url=self.base_url,
            search_text=search_text,
            max_results=max_results,
        )
        data = _get_response(url)
        return json.loads(data, object_hook=lambda d: Suburb(**d))

    def get_municipalities(self, province: Province) -> List[Municipality]:
        url = "{base_url}/GetMunicipalities/?Id={province}".format(
            base_url=self.base_url,
            province=province.id,
        )
        data = _get_response(url)
        return json.loads(data, object_hook=lambda d: Municipality(**d))

    def get_schedule(self, province: Province, suburb: Suburb, stage: Stage) -> Tuple[tuple]:
        schedule = []
        url = "{base_url}/GetScheduleM/{suburb_id}/{stage}/{province_id}/3252".format(
            base_url=self.base_url,
            suburb_id=suburb.id,
            province_id=province.id,
            stage=stage.value,
        )
        data = _get_response(url)
        soup = BeautifulSoup(data, "html.parser")
        days_soup = soup.find_all("div", attrs={"class": "scheduleDay"})

        now = datetime.datetime.now()
        for day in days_soup:
            date_soup = day.find("div", attrs={"class": "dayMonth"})
            date_str = date_soup.get_text().strip()
            date = datetime.datetime.strptime(date_str, "%a, %d %b")
            date = date.replace(year=now.year)

            time_soup = day.find_all("a")
            for time_tag in time_soup:
                start_str, end_str = time_tag.get_text().strip().split(" - ")
                start = datetime.datetime.strptime(start_str, "%H:%M")
                end = datetime.datetime.strptime(end_str, "%H:%M")
                schedule.append((
                    now.replace(month=date.month, day=date.day, hour=start.hour, minute=start.minute, second=0,
                                microsecond=0).strftime("%Y-%m-%d %H:%M"),
                    now.replace(month=date.month, day=date.day, hour=end.hour, minute=end.minute, second=0,
                                microsecond=0).strftime("%Y-%m-%d %H:%M")
                ))

        return schedule

    def get_schedule_area_info(self, suburb: Suburb) -> AreaInfo:
        url = "{base_url}/GetScheduleAreaInfo/?Id={suburb_id}".format(
            base_url=self.base_url,
            suburb_id=suburb.id,
        )
        data = _get_response(url)

        soup = BeautifulSoup(data, "html.parser")
        items = soup.find_all("div", attrs={"class": "areaInfoItem"})
        period = items[3].contents[2].strip().split("\xa0to\xa0")
        area_info = {
            "province": {
                "id": int(items[0].find("input", attrs={"id": "provinceId"}).get('value').strip()),
                "name": items[0].find("input", attrs={"id": "province"}).get('value').strip(),
            },
            "city": {
                "id": int(items[1].find("input", attrs={"id": "municipalityId"}).get('value').strip()),
                "name": items[1].find("input", attrs={"id": "municipality"}).get('value').strip(),
            },
            "suburb": {
                "id": int(items[2].find("input", attrs={"id": "suburbId"}).get('value').strip()),
                "name": items[2].find("input", attrs={"id": "suburbName"}).get('value').strip(),
            },
            "period": period,
        }
        return AreaInfo(**area_info)

    def get_stage(self) -> Stage:
        url = "{base_url}/GetStatus".format(base_url=self.base_url)
        data = _get_response(url)
        return Stage.from_status(data)


def _get_response(url: str) -> Any:
    try:
        logging.log(logging.DEBUG, "GET {url}".format(url=url))

        # r = get(url, verify=False, cert=None)

        retries = urllib3.Retry(connect=15, read=5, redirect=0, backoff_factor=1)
        with urllib3.PoolManager(retries=retries, ca_certs=certifi.where()) as conn:
            r = conn.request('GET', url)

        logging.log(logging.DEBUG, "Response: {url}".format(url=r.status))
        return r.data
    except Exception as e:
        logging.log(logging.ERROR, "Unable to get result from Eskom. {e}".format(e=e))



#      A python library for getting Load Shedding schedules from Eskom.
#      Copyright (C) 2021  Werner Pieterson
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
import logging
from datetime import datetime
from typing import Dict

from .eskom import Eskom, Province, Stage, Suburb


def get_area_schedule(provider: Eskom, province: Province, suburb: Suburb, cached: bool = False) -> Dict[Stage, list]:
    stage_schedule = {}

    if cached:
        try:
            cache_file = ".cache/{suburb_id}.json".format(suburb_id=suburb.id)
            with open(cache_file, "r") as cache:
                stage_schedule = json.loads(cache.read(), object_pairs_hook=lambda pairs: {Stage(int(k)).value: v for k, v in pairs})

            today = datetime.now().date()
            first = datetime.strptime(stage_schedule.get(Stage.STAGE_1.value)[0][0], "%Y-%m-%d %H:%M").date()
            if today > first:
                stage_schedule = {}
        except Exception as e:
            logging.log(logging.ERROR, "Unable to get result from cache. {e}".format(e=e))

    if not stage_schedule:
        for stage in range(Stage.STAGE_1.value, Stage.STAGE_5.value):
            stage_schedule[stage] = provider.get_schedule(province=province, suburb=suburb, stage=stage)

        if cached:
            cache_file = ".cache/{suburb_id}.json".format(suburb_id=suburb.id)
            with open(cache_file, "w") as cache:
                cache.write(json.dumps(stage_schedule))

    return stage_schedule


def get_schedule(provider: Eskom, province: Province, suburb: Suburb, stage: Stage = None, cached: bool = True) -> Dict[int, list]:
    if not stage:
        try:
            stage = provider.get_stage()
        except Exception as e:
            logging.log(logging.ERROR, "Unknown status".format(e=e))
        else:
            if stage == Stage.NO_LOAD_SHEDDING:
                raise Exception("Schedule not available for stage: {stage}".format(stage=Stage.NO_LOAD_SHEDDING))

    area_schedule = get_area_schedule(provider, province, suburb, cached=cached)
    return area_schedule.get(stage.value)


def list_to_dict(schedule: list) -> Dict:
    schedule_dict = {}
    now = datetime.now()
    for item in schedule:
        start = datetime.strptime(item[0], "%Y-%m-%d %H:%M")
        end = datetime.strptime(item[1], "%Y-%m-%d %H:%M")

        schedule_dict[start.strftime("%Y-%m-%d")] = (
            now.replace(month=start.month, day=start.day, hour=start.hour, minute=start.minute, microsecond=0).strftime(
                "%H:%M"),
            now.replace(month=end.month, day=end.day, hour=end.hour, minute=end.minute, microsecond=0).strftime(
                "%H:%M"),
        )
    return schedule_dict


if __name__ == "__main__":
    eskom = Eskom()
    suburb = "Milnerton"
    suburbs = eskom.find_suburbs(search_text=suburb)
    print("Searching suburbs for {suburb}: {suburbs}".format(suburb=suburb, suburbs=suburbs))
    schedule = eskom.get_schedule(province=suburbs[0].province, suburb=suburbs[0], stage=Stage.STAGE_2)
    print("Schedule for {suburb}: {schedule}".format(suburb=suburb, schedule=schedule))



#      A python library for getting Load Shedding schedules from Eskom.
#      Copyright (C) 2021  Werner Pieterson
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List

from .eskom import Eskom, AreaInfo, Province, Stage, Suburb, Municipality


class MockEskom(Eskom):
    def find_suburbs(self, search_text: str, max_results: int = 300) -> List[Suburb]:
        return [
            {
                'Id': 1058852,
                'MunicipalityName': 'City of Cape Town',
                'Name': 'Milnerton',
                'ProvinceName': 'Western Cape',
                'Total': 704
            }, {
                'Id': 1058853,
                'MunicipalityName': 'City of Cape Town',
                'Name': 'Milnerton Golf Course',
                'ProvinceName': 'Western Cape',
                'Total': 0
            }, {
                'Id': 1058854,
                'MunicipalityName': 'City of Cape Town',
                'Name': 'Milnerton Outlying',
                'ProvinceName': 'Western Cape',
                'Total': 0
            }, {
                'Id': 1058855,
                'MunicipalityName': 'City of Cape Town',
                'Name': 'Milnerton Ridge',
                'ProvinceName': 'Western Cape',
                'Total': 704
            }, {
                'Id': 1058856,
                'MunicipalityName': 'City of Cape Town',
                'Name': 'Milnerton SP',
                'ProvinceName': 'Western Cape',
                'Total': 2816
            }, {
                'Id': 1069144,
                'MunicipalityName': 'City of Cape Town',
                'Name': 'Milnerton SP 1',
                'ProvinceName': 'Western Cape',
                'Total': 3520
            }, {
                'Id': 1069145,
                'MunicipalityName': 'City of Cape Town',
                'Name': 'Milnerton SP 2',
                'ProvinceName': 'Western Cape',
                'Total': 700
            }
        ]

    def get_municipalities(self, province: Province) -> List[Municipality]:
        return [
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Beaufort West', 'Value': '336'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Bergrivier', 'Value': '337'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Bitou', 'Value': '338'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Breede Valley', 'Value': '339'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Cape Agulhas', 'Value': '340'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Cederberg', 'Value': '341'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'City of Cape Town', 'Value': '342'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Drakenstein', 'Value': '343'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'George', 'Value': '344'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Hessequa', 'Value': '345'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Kannaland', 'Value': '346'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Knysna', 'Value': '347'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Laingsburg', 'Value': '348'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Langeberg', 'Value': '349'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Matzikama', 'Value': '350'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Mossel Bay', 'Value': '351'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Oudtshoorn', 'Value': '352'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Overstrand', 'Value': '353'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Prince Albert', 'Value': '354'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Saldanha Bay', 'Value': '355'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Stellenbosch', 'Value': '356'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Swartland', 'Value': '357'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Swellendam', 'Value': '358'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Theewaterskloof', 'Value': '359'},
            {'Disabled': False, 'Group': None, 'Selected': False, 'Text': 'Witzenberg', 'Value': '360'}
        ]

    def get_schedule(self, province: Province, suburb: Suburb, stage: Stage) -> List[tuple]:
        return [
            ['2021-11-11 06:00', '2021-11-11 08:30'], ['2021-11-12 06:00', '2021-11-12 08:30'],
            ['2021-11-12 14:00', '2021-11-12 16:30'], ['2021-11-13 12:00', '2021-11-13 14:30'],
            ['2021-11-13 20:00', '2021-11-13 22:30'], ['2021-11-14 20:00', '2021-11-14 22:30'],
            ['2021-11-15 04:00', '2021-11-15 06:30'], ['2021-11-16 04:00', '2021-11-16 06:30'],
            ['2021-11-16 12:00', '2021-11-16 14:30'], ['2021-11-17 10:00', '2021-11-17 12:30'],
            ['2021-11-17 18:00', '2021-11-17 20:30'], ['2021-11-18 18:00', '2021-11-18 20:30'],
            ['2021-11-19 02:00', '2021-11-19 04:30'], ['2021-11-20 02:00', '2021-11-20 04:30'],
            ['2021-11-20 10:00', '2021-11-20 12:30'], ['2021-11-21 08:00', '2021-11-21 10:30'],
            ['2021-11-21 16:00', '2021-11-21 18:30'], ['2021-11-22 16:00', '2021-11-22 18:30'],
            ['2021-11-23 00:00', '2021-11-23 02:30'], ['2021-11-24 00:00', '2021-11-24 02:30'],
            ['2021-11-24 08:00', '2021-11-24 10:30'], ['2021-11-25 06:00', '2021-11-25 08:30'],
            ['2021-11-25 14:00', '2021-11-25 16:30'], ['2021-11-26 14:00', '2021-11-26 16:30'],
            ['2021-11-26 22:00', '2021-11-26 00:30'], ['2021-11-27 22:00', '2021-11-27 00:30'],
            ['2021-11-28 06:00', '2021-11-28 08:30'], ['2021-11-29 04:00', '2021-11-29 06:30'],
            ['2021-11-29 12:00', '2021-11-29 14:30'], ['2021-11-30 12:00', '2021-11-30 14:30'],
            ['2021-11-30 20:00', '2021-11-30 22:30'], ['2021-12-01 18:00', '2021-12-01 20:30'],
            ['2021-12-02 02:00', '2021-12-02 04:30'], ['2021-12-03 02:00', '2021-12-03 04:30'],
            ['2021-12-03 10:00', '2021-12-03 12:30'], ['2021-12-04 10:00', '2021-12-04 12:30'],
            ['2021-12-04 18:00', '2021-12-04 20:30'], ['2021-12-05 16:00', '2021-12-05 18:30'],
            ['2021-12-06 00:00', '2021-12-06 02:30'], ['2021-12-07 00:00', '2021-12-07 02:30'],
            ['2021-12-07 08:00', '2021-12-07 10:30'], ['2021-12-08 08:00', '2021-12-08 10:30'],
            ['2021-12-08 16:00', '2021-12-08 18:30'], ['2021-12-09 14:00', '2021-12-09 16:30'],
            ['2021-12-09 22:00', '2021-12-09 00:30']
        ]

    def get_schedule_area_info(self, suburb: Suburb) -> AreaInfo:
        return {
            'province': {'id': 9, 'name': 'Western Cape'},
            'city': {'id': 342, 'name': 'City of Cape Town'},
            'suburb': {'id': 1058852, 'name': 'Milnerton'},
            'period': ['11-11-2021', '08-12-2021']
        }

    def get_stage(self) -> Stage:
        return Stage.NO_LOAD_SHEDDING
